@@ -3,3 +3,7 @@
 from .doubles import *
 from .matchers import *
 from .internal import WrongApiUsage
+
+
+def set_default_behavior(double, func):
+    double._default_behavior = func
