@@ -19,7 +19,6 @@
 # Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 
-import itertools
 import threading
 import collections
 
@@ -48,7 +47,7 @@ class Constant(str):
 
 
 ANY_ARG = Constant('ANY_ARG')
-IMPOSSIBLE = Constant('IMPOSSIBLE')
+UNSPECIFIED = Constant('UNSPECIFIED')
 
 
 def add_indent(text, indent=0):
@@ -107,13 +106,13 @@ class Method(Observable):
         return retval
 
     def _create_invocation(self, args, kargs):
-        return Invocation.from_args(self.double, self.name, args, kargs)
+        return Invocation._from_args(self.double, self.name, args, kargs)
 
     @property
     def calls(self):
         if not isinstance(self.double, SpyBase):
             raise WrongApiUsage("Only Spy derivates store invocations")
-        return [x.context for x in self.double._get_invocations_to(self.name)]
+        return [x._context for x in self.double._get_invocations_to(self.name)]
 
     def _was_called(self, context, times):
         invocation = Invocation(self.double, self.name, context)
@@ -122,13 +121,13 @@ class Method(Observable):
     def describe_to(self, description):
         pass
 
-    def show(self, indent=0):
+    def _show(self, indent=0):
         return add_indent(self, indent)
 
     def __repr__(self):
         return "%s.%s" % (self.double._classname(), self.name)
 
-    def show_history(self):
+    def _show_history(self):
         method = "method '%s.%s'" % (self.double._classname(), self.name)
         invocations = self.double._get_invocations_to(self.name)
         if not invocations:
@@ -164,33 +163,34 @@ def func_raising(e):
 @total_ordering
 class Invocation(object):
     def __init__(self, double, name, context=None):
-        self.double = double
-        self.name = name
-        self.context = context or InvocationContext()
-        self.delegate = func_returning(None)
+        self._double = double
+        self._name = name
+        self._context = context or InvocationContext()
+        self._context.signature = double._proxy.get_signature(name)
+        self.__delegate = func_returning(None)
 
     @classmethod
-    def from_args(cls, double, name, args=(), kargs={}):
+    def _from_args(cls, double, name, args=(), kargs={}):
         return Invocation(double, name, InvocationContext(*args, **kargs))
 
     def delegates(self, delegate):
         if isinstance(delegate, collections.Callable):
-            self.delegate = delegate
+            self.__delegate = delegate
             return
 
         try:
-            self.delegate = iter(delegate).next
+            self.__delegate = iter(delegate).next
         except TypeError:
             reason = "delegates() must be called with callable or iterable instance (got '%s' instead)" % delegate
             raise WrongApiUsage(reason)
 
     def returns(self, value):
-        self.context.retval = value
+        self._context.retval = value
         self.delegates(func_returning(value))
         return self
 
     def returns_input(self):
-        if not self.context.args:
+        if not self._context.args:
             raise TypeError("%s has no input args" % self)
 
         self.delegates(func_returning_input(self))
@@ -204,58 +204,63 @@ class Invocation(object):
             raise WrongApiUsage("times must be >= 1. Use is_not(called()) for 0 times")
 
         for i in range(1, n):
-            self.double._manage_invocation(self)
+            self._double._manage_invocation(self)
 
-    # FIXME: rename to apply_stub?
-    def perform(self, actual_invocation):
-        context = actual_invocation.context
-        return self.delegate(*context.args, **context.kargs)
+    def _apply_stub(self, actual_invocation):
+        return actual_invocation._context.apply_on(self.__delegate)
 
-    def apply_on_collaborator(self):
-        return self.double._proxy.perform_invocation(self)
+    def _apply_on_collaborator(self):
+        return self._double._proxy.perform_invocation(self)
 
     def __eq__(self, other):
-        return self.double._proxy.same_method(self.name, other.name) and \
-            self.context.matches(other.context)
+        return self._double._proxy.same_method(self._name, other._name) and \
+            self._context.matches(other._context)
 
     def __lt__(self, other):
-        return any([self.name < other.name,
-                    self.context < other.context])
+        return any([self._name < other._name,
+                    self._context < other._context])
 
     def __repr__(self):
-        return "%s.%s%s" % (self.double._classname(), self.name, self.context)
+        return "%s.%s%s" % (self._double._classname(), self._name, self._context)
 
-    def show(self, indent=0):
+    def _show(self, indent=0):
         return add_indent(self, indent)
 
 
+ANY_ARG_MUST_BE_LAST = "ANY_ARG must be the last positional argument. "
+ANY_ARG_WITHOUT_KARGS = "Keyword arguments are not allowed if ANY_ARG is given. "
+ANY_ARG_CAN_BE_KARG = "ANY_ARG is not allowed as keyword value. "
+ANY_ARG_DOC = "See http://goo.gl/R6mOt"
+
 @total_ordering
 class InvocationContext(object):
     def __init__(self, *args, **kargs):
         self.update_args(args, kargs)
         self.retval = None
+        self.signature = None
+        self.check_some_args = False
 
     def update_args(self, args, kargs):
+        self._check_ANY_ARG_sanity(args, kargs)
         self.args = args
         self.kargs = kargs
 
-    def matches(self, other):
+    def _check_ANY_ARG_sanity(self, args, kargs):
         try:
-            if self._assert_args_match(self.args, other.args) is ANY_ARG:
-                return True
+            if args.index(ANY_ARG) != len(args)-1:
+                raise WrongApiUsage(ANY_ARG_MUST_BE_LAST + ANY_ARG_DOC)
 
-            self._assert_kargs_match(self.kargs, other.kargs)
-            return True
-        except AssertionError:
-            return False
+            if kargs:
+                raise WrongApiUsage(ANY_ARG_WITHOUT_KARGS + ANY_ARG_DOC)
+        except ValueError:
+            pass
 
-    @classmethod
-    def _assert_args_match(cls, args1, args2):
-        for a, b in itertools.izip_longest(args1, args2, fillvalue=IMPOSSIBLE):
-            if ANY_ARG in [a, b]:
-                return ANY_ARG
+        if ANY_ARG in kargs.values():
+            raise WrongApiUsage(ANY_ARG_CAN_BE_KARG + ANY_ARG_DOC)
 
-            cls._assert_values_match(a, b)
+
+    def apply_on(self, method):
+        return method(*self.args, **self.kargs)
 
     @classmethod
     def _assert_kargs_match(cls, kargs1, kargs2):
@@ -270,8 +275,54 @@ class InvocationContext(object):
 
         hamcrest.assert_that(a, hamcrest.is_(b))
 
-    def __eq__(self, other):
-        return self.matches(other)
+    def copy(self):
+        retval = InvocationContext(*self.args, **self.kargs)
+        retval.signature = self.signature
+        return retval
+
+    def replace_ANY_ARG(self, actual):
+        try:
+            index = self.args.index(ANY_ARG)
+        except ValueError:
+            return self
+
+        retval = self.copy()
+        args = list(self.args[0:index])
+        args.extend([hamcrest.anything()] * (len(actual.args) - index))
+        retval.args = tuple(args)
+        retval.kargs = actual.kargs.copy()
+        return retval
+
+    def matches(self, other):
+        if ANY_ARG in self.args:
+            matcher, actual = self, other
+        else:
+            matcher, actual = other, self
+
+        matcher = matcher.replace_ANY_ARG(actual)
+
+        if matcher.check_some_args:
+            matcher.kargs = self.add_unspecifed_args(matcher)
+
+        matcher_call_args = matcher.signature.get_call_args(matcher)
+        actual_call_args = actual.signature.get_call_args(actual)
+
+        try:
+            self._assert_kargs_match(matcher_call_args, actual_call_args)
+            return True
+        except AssertionError:
+            return False
+
+    def add_unspecifed_args(self, context):
+        arg_spec = context.signature.get_arg_spec()
+        if arg_spec is None:
+            raise WrongApiUsage(
+                'free spies does not support the with_some_args() matcher')
+
+        keys = arg_spec.args
+        retval = dict((k, hamcrest.anything()) for k in keys)
+        retval.update(context.kargs)
+        return retval
 
     def __lt__(self, other):
         if ANY_ARG in other.args or self.args < other.args:
@@ -320,18 +371,18 @@ class InvocationFormatter(object):
 
 class PropertyInvocation(Invocation):
     def __eq__(self, other):
-        return self.name == other.name
+        return self._name == other._name
 
 
 class PropertyGet(PropertyInvocation):
     def __init__(self, double, name):
         super(PropertyGet, self).__init__(double, name)
 
-    def apply_on_collaborator(self):
-        return getattr(self.double._proxy.collaborator, self.name)
+    def _apply_on_collaborator(self):
+        return getattr(self._double._proxy.collaborator, self._name)
 
     def __repr__(self):
-        return "get %s.%s" % (self.double._classname(), self.name)
+        return "get %s.%s" % (self._double._classname(), self._name)
 
 
 class PropertySet(PropertyInvocation):
@@ -340,12 +391,12 @@ class PropertySet(PropertyInvocation):
         param = InvocationContext(value)
         super(PropertySet, self).__init__(double, name, param)
 
-    def apply_on_collaborator(self):
-        return setattr(self.double._proxy.collaborator, self.name, self.value)
+    def _apply_on_collaborator(self):
+        return setattr(self._double._proxy.collaborator, self._name, self.value)
 
     def __repr__(self):
-        return "set %s.%s to %s" % (self.double._classname(),
-                                    self.name, self.value)
+        return "set %s.%s to %s" % (self._double._classname(),
+                                    self._name, self.value)
 
 
 def property_factory(double, key):
