@@ -2,7 +2,7 @@
 
 # doublex
 #
-# Copyright © 2012 David Villa Alises
+# Copyright © 2012,2013 David Villa Alises
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -32,6 +32,7 @@ from hamcrest.library.object.hasproperty import *
 from hamcrest.library.number.ordering_comparison import *
 
 from doublex import *
+from doublex.matchers import MatcherRequiredError
 from doublex.internal import InvocationContext
 
 
@@ -93,16 +94,22 @@ class FreeStubTests(TestCase):
         with self.stub:
             self.stub.foo().returns(True)
 
-        assert_that(self.stub.bar(), is_(None))
+        assert_that(self.stub.unknown(), is_(None))
+
+    def test_returns_input(self):
+        with Stub() as stub:
+            stub.foo(1).returns_input()
+
+        assert_that(stub.foo(1), is_(1))
 
     def test_raises(self):
         with self.stub:
-            self.stub.foo().raises(KeyError)
+            self.stub.foo(2).raises(SomeException)
 
         try:
-            self.stub.foo()
-            self.fail("It should raise KeyError")
-        except KeyError:
+            self.stub.foo(2)
+            self.fail("It should raise SomeException")
+        except SomeException:
             pass
 
 
@@ -144,19 +151,33 @@ class StubTests(TestCase):
             assert_that(str(e), contains_string(expected))
 
     # bitbucket issue #6
-    def test_keyworked_or_positional(self):
+    def test_keyword_or_positional(self):
         with self.stub:
             self.stub.kwarg_method(1).returns(1000)
-            self.stub.kwarg_method(key_param=2).returns(2000)
+            self.stub.kwarg_method(2).returns(2000)
+            self.stub.kwarg_method(key_param=6).returns(6000)
 
         assert_that(self.stub.kwarg_method(1), is_(1000))
-        assert_that(self.stub.kwarg_method(key_param=2), is_(2000))
+        assert_that(self.stub.kwarg_method(2), is_(2000))
+        assert_that(self.stub.kwarg_method(key_param=6), is_(6000))
+        assert_that(self.stub.kwarg_method(key_param=6), is_(6000))
+
+    # FIXME: new on 1.7
+    def test_keyworked_or_positional_are_equivalent(self):
+        with self.stub:
+            self.stub.kwarg_method(1).returns(1000)
+            self.stub.kwarg_method(key_param=6).returns(6000)
+
+        assert_that(self.stub.kwarg_method(1), is_(1000))
+        assert_that(self.stub.kwarg_method(key_param=1), is_(1000))
+        assert_that(self.stub.kwarg_method(6), is_(6000))
+        assert_that(self.stub.kwarg_method(key_param=6), is_(6000))
 
     def test_returning_tuple(self):
         with self.stub:
             self.stub.hello().returns((3, 4))
 
-        assert_that(self.stub.hello(), (3, 4))
+        assert_that(self.stub.hello(), is_((3, 4)))
 
 
 class AccessingActualAttributes(TestCase):
@@ -303,7 +324,7 @@ class FreeSpyTests(TestCase):
 #        assert_that(spy.mixed_method, called().with_args(key_param=True))
 
 
-class SpyCallsTests(TestCase):
+class Spy_calls_tests(TestCase):
     def test_list_recorded_calls(self):
         class Collaborator:
             def method(self, *args, **kargs):
@@ -344,7 +365,7 @@ class SpyCallsTests(TestCase):
 
 class SpyTests(TestCase):
     def setUp(self):
-        self.spy = Spy(Collaborator())
+        self.spy = Spy(Collaborator)
 
     def test_from_instance(self):
         spy = Spy(Collaborator())
@@ -374,6 +395,11 @@ class SpyTests(TestCase):
     def test_create_from_newstyle_class(self):
         Spy(ObjCollaborator)
 
+    def test_wrong_call_args(self):
+        self.spy.hello()
+        with self.assertRaises(TypeError):
+            assert_that(self.spy.hello, called().with_args('some'))
+
 
 class BuiltinSpyTests(TestCase):
     def test_builtin_method(self):
@@ -575,19 +601,19 @@ class DisplayResultsTests(TestCase):
             self.spy.method_one(ANY_ARG).returns(2)
 
     def test_empty_spy_stub_method(self):
-        assert_that(self.empty_spy.foo.show_history(),
-                    "method 'Spy.foo' never invoked")
+        assert_that(self.empty_spy.foo._show_history(),
+                    reason="method 'Spy.foo' never invoked")
 
     def test_spy_stub_method(self):
-        assert_that(self.spy.method_one.show_history(),
-                    "method 'Collaborator.method_one' never invoked")
+        assert_that(self.spy.method_one._show_history(),
+                    reason="method 'Collaborator.method_one' never invoked")
 
     def test_empty_spy_stub_method_invoked(self):
         self.empty_spy.foo()
         expected = [
             "method 'Spy.foo' was invoked",
             "foo()"]
-        assert_that(self.empty_spy.foo.show_history(),
+        assert_that(self.empty_spy.foo._show_history(),
                     string_contains_in_order(*expected))
 
     def test_spy_stub_method_invoked(self):
@@ -595,7 +621,7 @@ class DisplayResultsTests(TestCase):
         expected = [
             "method 'Collaborator.method_one' was invoked",
             'method_one(1)']
-        assert_that(self.spy.method_one.show_history(),
+        assert_that(self.spy.method_one._show_history(),
                     string_contains_in_order(*expected))
 
     def test_empty_spy_non_stubbed_method_invoked(self):
@@ -603,7 +629,7 @@ class DisplayResultsTests(TestCase):
         expected = [
             "method 'Spy.bar' was invoked",
             "bar(1, 3.0, 'text', key1='text', key2=[1, 2])"]
-        assert_that(self.empty_spy.bar.show_history(),
+        assert_that(self.empty_spy.bar._show_history(),
                     string_contains_in_order(*expected))
 
     def test_spy_several_invoked_same_method(self):
@@ -611,7 +637,7 @@ class DisplayResultsTests(TestCase):
         self.spy.mixed_method(8, False)
 
         expected = "method 'Collaborator.mixed_method' was invoked"
-        assert_that(self.spy.mixed_method.show_history(),
+        assert_that(self.spy.mixed_method._show_history(),
                     contains_string(expected))
 
 
@@ -688,6 +714,11 @@ class ANY_ARG_StubTests(TestCase):
         assert_that(self.stub.foo(1, 2, 3), is_(True))
         assert_that(self.stub.foo(1, key1='a'), is_(True))
 
+    def test_ANY_ARG_must_be_last_positional_argument(self):
+        with self.assertRaises(WrongApiUsage):
+            with self.stub:
+                self.stub.method(1, ANY_ARG, 3).returns(True)
+
 
 class ANY_ARG_SpyTests(TestCase):
     def setUp(self):
@@ -696,10 +727,13 @@ class ANY_ARG_SpyTests(TestCase):
     def test_no_args(self):
         self.spy.foo()
         assert_that(self.spy.foo, called().with_args(ANY_ARG))
+        assert_that(self.spy.foo, never(called().with_args(anything())))
 
     def test_one_arg(self):
         self.spy.foo(1)
+        assert_that(self.spy.foo, called())
         assert_that(self.spy.foo, called().with_args(ANY_ARG))
+        assert_that(self.spy.foo, called().with_args(anything()))
 
     def test_one_karg(self):
         self.spy.foo(key='val')
@@ -724,6 +758,43 @@ class ANY_ARG_SpyTests(TestCase):
         assert_that(self.spy.foo, called().times(4))
         assert_that(self.spy.foo, called().with_args(ANY_ARG).times(4))
 
+    # issue 9
+    def test_ANY_ARG_forbbiden_as_keyword_value(self):
+        person = Spy()
+        person.set_info(name="John", surname="Doe")
+
+        assert_that(person.set_info,
+                    called().with_args(name=anything(), surname="Doe"))
+
+        with self.assertRaises(WrongApiUsage):
+            assert_that(person.set_info,
+                        called().with_args(name=ANY_ARG, surname="Doe"))
+
+    def test_ANY_ARG_must_be_last_positional_argument(self):
+        self.spy.method(1, 2, 3)
+
+        with self.assertRaises(WrongApiUsage):
+            assert_that(self.spy.method,
+                        called().with_args(1, ANY_ARG, 3))
+
+    def test_ANY_ARG_must_be_last_positional_argument_with_xarg(self):
+        self.spy.method(1, 2, 3, name='Bob')
+
+        with self.assertRaises(WrongApiUsage):
+            assert_that(self.spy.method,
+                        called().with_args(1, ANY_ARG, name='Bob'))
+
+    def test_ANY_ARG_must_be_last_positional_argument__restricted_spy(self):
+        spy = Spy(Collaborator)
+
+        with self.assertRaises(WrongApiUsage):
+            assert_that(spy.two_args_method,
+                        called().with_args(ANY_ARG, 2))
+
+        with self.assertRaises(WrongApiUsage):
+            assert_that(spy.three_args_method,
+                        called().with_args(1, ANY_ARG, 3))
+
 
 class MatcherTests(TestCase):
     def setUp(self):
@@ -776,7 +847,7 @@ class MatcherTests(TestCase):
         assert_that(self.spy.foo, is_not(called().with_args(5)))                 # = 0 times
         assert_that(self.spy.foo, called().with_args().times(1))                 # = 1
         assert_that(self.spy.foo, called().with_args(anything()))                # > 0
-        assert_that(self.spy.foo, called().with_args(anything()).times(4))       # = 4
+        assert_that(self.spy.foo, called().with_args(ANY_ARG).times(4))          # = 4
         assert_that(self.spy.foo, called().with_args(1).times(2))                # = 2
         assert_that(self.spy.foo, called().with_args(1).times(greater_than(1)))  # > 1
         assert_that(self.spy.foo, called().with_args(1).times(less_than(5)))     # < 5
@@ -804,6 +875,10 @@ class MatcherTests(TestCase):
         assert_that(self.spy.m3, called().with_args(greater_than(1)))
         assert_that(self.spy.m6, called().with_args(name=contains_string("doe")))
 
+    # new on 1.7
+    def test_assert_that_requires_a_matcher(self):
+        self.assertRaises(MatcherRequiredError, assert_that, self.spy.m1, True)
+
 
 class StubObserverTests(TestCase):
     def setUp(self):
@@ -915,7 +990,7 @@ class MimicTests(TestCase):
     def test_mimic_spy_DOES_inherit_collaborator_superclasses(self):
         spy = Mimic(Spy, self.B)
         for cls in [self.B, self.A, Spy, Stub, object]:
-            assert_that(isinstance(spy, cls), cls)
+            assert_that(spy, instance_of(cls))
 
     def test_mimic_stub_works(self):
         stub = Mimic(Stub, self.B)
@@ -1139,6 +1214,152 @@ class AsyncTests(TestCase):
         assert_that(spy.write, called().async(timeout=1))
 
 
+# FIXME: new on 1.7
+class with_some_args_matcher_tests(TestCase):
+    def test_one_arg(self):
+        spy = Spy(Collaborator)
+        spy.mixed_method(5)
+        assert_that(spy.mixed_method, called().with_args(5))
+        assert_that(spy.mixed_method, called().with_args(arg1=5))
+
+    def test_two_arg(self):
+        spy = Spy(Collaborator)
+        spy.two_args_method(5, 10)
+        assert_that(spy.two_args_method, called().with_args(5, 10))
+        assert_that(spy.two_args_method, called().with_args(arg1=5, arg2=10))
+        assert_that(spy.two_args_method, called().with_some_args(arg1=5))
+        assert_that(spy.two_args_method, called().with_some_args(arg2=10))
+        assert_that(spy.two_args_method, called().with_some_args())
+
+    def test_free_spy(self):
+        spy = Spy()
+        spy.foo(1, 3)
+
+        with self.assertRaises(WrongApiUsage):
+            assert_that(spy.foo, called().with_some_args())
+
+
+# FIXME: new on 1.7
+class Stub_default_behavior_tests(TestCase):
+    def test_set_return_globally(self):
+        StubClone = Stub._clone_class()
+        set_default_behavior(StubClone, method_returning(20))
+        stub = StubClone()
+
+        assert_that(stub.unknown(), is_(20))
+
+    def test_set_exception_globally(self):
+        StubClone = Stub._clone_class()
+        set_default_behavior(StubClone, method_raising(SomeException))
+        stub = StubClone()
+
+        with self.assertRaises(SomeException):
+            stub.unknown()
+
+    def test_set_return_by_instance(self):
+        stub = Stub()
+        set_default_behavior(stub, method_returning(20))
+
+        assert_that(stub.unknown(), is_(20))
+
+    def test_set_exception_by_instance(self):
+        stub = Stub()
+        set_default_behavior(stub, method_raising(SomeException))
+
+        with self.assertRaises(SomeException):
+            stub.unknown()
+
+    def test_restricted_stub(self):
+        stub = Stub(Collaborator)
+        set_default_behavior(stub, method_returning(30))
+        with stub:
+            stub.hello().returns(1000)
+
+        assert_that(stub.something(), is_(30))
+        assert_that(stub.hello(), is_(1000))
+
+
+# FIXME: new on 1.7
+class Spy_default_behavior_tests(TestCase):
+    def test_set_return_globally(self):
+        SpyClone = Spy._clone_class()
+        set_default_behavior(SpyClone, method_returning(20))
+        spy = SpyClone()
+
+        assert_that(spy.unknown(7), is_(20))
+
+        assert_that(spy.unknown, called().with_args(7))
+        assert_that(spy.unknown, never(called().with_args(9)))
+
+    def test_set_return_by_instance(self):
+        spy = Spy()
+        set_default_behavior(spy, method_returning(20))
+
+        assert_that(spy.unknown(7), is_(20))
+
+        assert_that(spy.unknown, called().with_args(7))
+
+
+# FIXME: new on 1.7
+class ProxySpy_default_behavior_tests(TestCase):
+    def test_this_change_proxyspy_default_behavior(self):
+        spy = ProxySpy(Collaborator())
+        assert_that(spy.hello(), is_("hello"))
+
+        set_default_behavior(spy, method_returning(40))
+        assert_that(spy.hello(), is_(40))
+
+
+# FIXME: new on tip
+class orphan_methods_tests(TestCase):
+    def setUp(self):
+        self.obj = Collaborator()
+
+    def test_stub_method(self):
+        with Stub() as stub:
+            stub.method(1).returns(100)
+            stub.method(2).returns(200)
+
+        self.obj.foo = stub.method
+
+        assert_that(self.obj.foo(0), is_(None))
+        assert_that(self.obj.foo(1), is_(100))
+        assert_that(self.obj.foo(2), is_(200))
+
+    def test_spy_method(self):
+        with Spy() as spy:
+            spy.method(1).returns(100)
+            spy.method(2).returns(200)
+            spy.method(3).raises(SomeException)
+
+        self.obj.foo = spy.method
+
+        assert_that(self.obj.foo(0), is_(None))
+        assert_that(self.obj.foo(1), is_(100))
+        assert_that(self.obj.foo(2), is_(200))
+        with self.assertRaises(SomeException):
+            self.obj.foo(3)
+
+        assert_that(self.obj.foo, called().times(4))
+        assert_that(spy.method, called().times(4))
+
+#    def test_spy_method__brief_method(self):
+#        with method() as self.obj.foo:
+#            self.obj.foo().returns(100)
+#            self.obj.foo(2).returns(200)
+#
+#        assert_that(self.obj.foo(), is_(100))
+#        assert_that(self.obj.foo(2), is_(200))
+#        assert_that(self.obj.foo(3), is_(None))
+#
+#        assert_that(self.obj.foo, called().times(3))
+#        assert_that(spy.method, called().times(3))
+
+
+class SomeException(Exception):
+    pass
+
+
 class Observer(object):
     def __init__(self):
         self.state = None
@@ -1188,6 +1409,9 @@ class Collaborator:
     def two_args_method(self, arg1, arg2):
         return arg1 + arg2
 
+    def three_args_method(self, arg1, arg2, arg3):
+        return arg1 + arg2 + arg3
+
     def kwarg_method(self, key_param=False):
         return key_param
 
